'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global DomainSet, Filter, FilterOptions, getSettings, ElementTypes */
/* eslint-disable no-bitwise */

// A single filter rule.
const Filter = function () {
  Filter.lastId += 1;
  this.id = Filter.lastId;
};

// Filters that block by CSS selector.
const SelectorFilter = function (text) {
  Filter.call(this); // call base constructor

  const parts = text.match(/(^.*?)#@?#(.+$)/);
  this.domains = Filter.toDomainSet(parts[1], ',');
  [,, this.selector] = parts;
  if (typeof getSettings() === 'object' && getSettings().show_advanced_options) {
    this.text = text;
  }
};

// Filters that block by CSS selector.
const ElemHideEmulationFilter = function (text) {
  Filter.call(this); // call base constructor

  const parts = text.match(/(^.*?)#\?#(.+$)/);
  this.domains = Filter.toDomainSet(parts[1], ',');
  [,, this.selector] = parts;
  this.text = text;
};

// Filters that block by URL regex or substring.
const PatternFilter = function () {
  Filter.call(this); // call base constructor
};

/**
 * Base class for content filters
 * @param {string} text
 * @param {string} [domains] Host names or domains the filter should be
 *                           restricted to
 * @param {string} body      The body of the filter
 */
const ContentFilter = function (text, domains, body) {
  Filter.call(this); // call base constructor
  this.body = body;
  this.domains = Filter.toDomainSet(domains, ',');
  this.text = text;
};

/**
 * Class for snippet filters
 * @param {string} text
 * @param {string} [domains]
 * @param {string} script    Script that should be executed
 * @constructor
 * @augments ContentFilter
 */
const SnippetFilter = function (text, domains, script) {
  Filter.call(this); // call base constructor
  this.body = script;
  this.domains = Filter.toDomainSet(domains, ',');
  this.text = text;
};

Filter.lastId = 0;

// Maps filter text to Filter instances.  This is important, as it allows
// us to throw away and rebuild the FilterSet at will.
// Will be cleared after a fixed time interval
Filter.cache = {};

// Return a Filter instance for the given filter text.
// Throw an exception if the filter is invalid.
Filter.fromText = function filterFromText(text) {
  const { cache } = Filter;
  if (!(text in cache)) {
    if (Filter.isSelectorFilter(text)) {
      cache[text] = new SelectorFilter(text);
    } else if (Filter.isAdvancedSelectorFilter(text)) {
      cache[text] = new ElemHideEmulationFilter(text);
    } else if (Filter.isContentFilter(text)) {
      cache[text] = ContentFilter.fromText(text);
    } else {
      cache[text] = PatternFilter.fromText(text);
    }
  }

  return cache[text];
};

// Test if pattern#@#pattern or pattern##pattern
Filter.isSelectorFilter = function checkSelectorFilter(text) {
  // This returns true for both hiding rules as hiding whitelist rules
  // This means that you'll first have to check if something is an excluded rule
  // before checking this, if the difference matters.
  return /#@?#./.test(text);
};

Filter.isSelectorExcludeFilter = function checkSelectorExcludeFilter(text) {
  return /#@#./.test(text);
};

// Test if pattern#?#pattern
Filter.isAdvancedSelectorFilter = function checkAdvancedSelectorFilter(text) {
  return /#\?#./.test(text);
};

/**
 * Regular expression that content filters should match
 * @type {RegExp}
 */
Filter.contentRegExp = /^([^/*|@"!]*?)#([@?$])?#(.+)$/;
/**
 * test if pattern# @ | ? | $ #pattern
 */
Filter.isContentFilter = function checkContentFilter(text) {
  return Filter.contentRegExp.test(text);
};

/**
 * Regular expression that snippet filters should match
 * @type {RegExp}
 */
Filter.snippetRegExp = /^([^/*|@"!]*?)#$#(.+)$/;
/**
 * test if pattern#$#pattern
 */
Filter.isSnippetFilter = function checkSnippetFilter(text) {
  return Filter.snippetRegExp.test(text);
};

Filter.isWhitelistFilter = function checkWhitelistFilter(text) {
  return /^@@/.test(text);
};

/**
 * Regular expression that content filters should match
 * @type {RegExp}
 */
Filter.contentRegExp = /^([^/*|@"!]*?)#([@?$])?#(.+)$/;
/**
 * test if pattern# @ | ? | $ #pattern
 */
Filter.isContentFilter = function checkContentFilter(text) {
  return Filter.contentRegExp.test(text);
};

Filter.isComment = function checkComment(text) {
  return text.length === 0
      || text[0] === '!'
      || (/^\[adblock/i.test(text))
      || (/^\(adblock/i.test(text));
};

/**
 * Regular expression that element hiding filters should match
 */
Filter.elemhideRegExp = /^([^/*|@"!]*?)#([@?])?#(.+)$/;
/**
 * Regular expression that old advance hiding filters should match
 */
Filter.oldAdvanceSyntaxRegExp = /\[-abp-properties=(["'])([^"']+)\1\]/;
/**
 * Regular expression that matches an invalid Content Security Policy
 * @type {RegExp}
 */
Filter.invalidCSPRegExp = /(base-uri|referrer|report-to|report-uri|upgrade-insecure-requests)\b/i;

Filter.optionsRegex = /\$~?[\w-]+(?:=[^,]+)?(?:,~?[\w-]+(?:=[^,]+)?)*$/;

// Convert a comma-separated list of domain includes and excludes into a
// DomainSet.
Filter.toDomainSet = function filterToDomainSet(domainText, divider) {
  const domains = domainText.split(divider);

  const data = {};
  data[DomainSet.ALL] = true;

  /* eslint-disable-next-line eqeqeq */
  if (domains == '') {
    return new DomainSet(data);
  }

  let domainsLength = domains.length;
  while (domainsLength > 0) {
    domainsLength -= 1;
    const domain = domains[domainsLength];
    if (domain[0] === '~') {
      data[domain.substring(1)] = false;
    } else {
      data[domain] = true;
      data[DomainSet.ALL] = false;
    }
  }

  return new DomainSet(data);
};

// If !|excludeFilters|, returns filter.
// Otherwise, returns a new SelectorFilter that is the combination of
// |filter| and each selector exclusion filter in the given list.
SelectorFilter.merge = function selectorFilterMerge(filter, excludeFilters) {
  if (!excludeFilters) {
    return filter;
  }

  const domains = filter.domains.clone();
  let excludeFiltersLength = excludeFilters.length;
  while (excludeFiltersLength > 0) {
    excludeFiltersLength -= 1;
    domains.subtract(excludeFilters[excludeFiltersLength].domains);
  }

  const result = new SelectorFilter('_##_');
  result.selector = filter.selector;
  if (filter.text) {
    result.text = filter.text;
  }
  result.domains = domains;

  return result;
};

SelectorFilter.prototype = {

  // Inherit from Filter.
  __proto__: Filter.prototype,
};

ElemHideEmulationFilter.prototype = {

  // Inherit from Filter - SelectorFilter.
  __proto__: SelectorFilter.prototype,
};

// Data is [rule text, allowed element types, options].
PatternFilter.fromData = function patternFilterFromData(data) {
  const result = new PatternFilter();
  result.rule = new RegExp(data[0]);
  [, result.allowedElementTypes, result.options] = data;
  const newData = {};
  newData[DomainSet.ALL] = true;
  result.domains = new DomainSet(newData);
  return result;
};

// Text is the original filter text of a blocking or whitelist filter.
// Throws an exception if the rule is invalid.
PatternFilter.fromText = function patternFilterFromText(text) {
  const data = PatternFilter.parseRule(text);

  const result = new PatternFilter();
  result.domains = Filter.toDomainSet(data.domainText, '|');
  result.allowedElementTypes = data.allowedElementTypes;
  result.options = data.options;
  result.rule = data.rule;
  result.key = data.key;
  result.cspText = data.cspText;
  result.rewrite = data.rewrite;

  if (typeof getSettings() === 'object' && getSettings().show_advanced_options) {
    result.text = text;
  }
  return result;
};

// Updates the |allowedElementTypes| for filter based on inverted arg
// returns the updated |allowedElementTypes|
function setAllowedElementTypes(allowedElementTypesArg, inverted, newType) {
  let allowedElementTypes = allowedElementTypesArg;
  if (inverted) {
    if (allowedElementTypes === undefined) {
      allowedElementTypes = ElementTypes.DEFAULTTYPES;
    }
    allowedElementTypes &= ~newType;
  } else {
    if (allowedElementTypes === undefined) {
      allowedElementTypes = ElementTypes.NONE;
    }
    allowedElementTypes |= newType;
  }
  return allowedElementTypes;
}

// Return a { rule, domainText, allowedElementTypes } object
// for the given filter text.  Throws an exception if the rule is invalid.
PatternFilter.parseRule = function patternFilterParseRule(text) {
  const result = {
    domainText: '',

    // TODO: when working on this code again, consider making options a
    // dictionary with boolean values instead of a bitset. This would
    // - make more sense, because these options are only checked individually
    // - collapse the two bitwise checks in Filter.matches into a single
    // boolean compare
    options: FilterOptions.NONE,
  };

  const optionsText = text.match(Filter.optionsRegex);
  let allowedElementTypes;
  let rule;
  let options;
  if (!optionsText) {
    rule = text;
    options = [];
  } else {
    options = optionsText[0].substring(1).toLowerCase().split(',');
    rule = text.replace(optionsText[0], '');
  }

  for (let i = 0; i < options.length; i += 1) {
    let option = options[i];

    if (/^domain=/.test(option)) {
      result.domainText = option.substring(7);
    } else {
      const inverted = (option[0] === '~');
      if (inverted) {
        option = option.substring(1);
      }
      const originalOptionText = option;
      option = option.replace(/-/, '_');

      if (option === 'object_subrequest') {
        option = 'object';
      }

      // 'background' is a synonym for 'image'.
      if (option === 'background') {
        option = 'image';
      }

      if (option in ElementTypes) { // this option is a known element type
        allowedElementTypes = setAllowedElementTypes(
          allowedElementTypes,
          inverted, ElementTypes[option],
        );
      } else if (option === 'third_party') {
        result.options
            |= (inverted ? FilterOptions.FIRSTPARTY : FilterOptions.THIRDPARTY);
      } else if (option === 'match_case') {
        // doesn't have an inverted function
        result.options |= FilterOptions.MATCHCASE;
      } else if (/^csp=/.test(originalOptionText)) {
        const cspText = originalOptionText.substring(4);
        if (cspText && Filter.invalidCSPRegExp.test(cspText)) {
          throw new Error(`invalid CSP option '${cspText}' in filter '${text}'`);
        }
        result.cspText = cspText;
        allowedElementTypes = setAllowedElementTypes(
          allowedElementTypes,
          inverted, ElementTypes.csp,
        );
      } else if (/^rewrite=/.test(originalOptionText)) {
        result.rewrite = originalOptionText.substring(8);
        allowedElementTypes = setAllowedElementTypes(
          allowedElementTypes,
          inverted, ElementTypes.DEFAULTTYPES,
        );
        // For security reasons, never match $rewrite filters
        // against requests that might load any code to be executed.
        allowedElementTypes &= ~(ElementTypes.script
                           | ElementTypes.subdocument
                           | ElementTypes.object
                           | ElementTypes.object_subrequest);
      } else if (option === 'collapse') {

        // We currently do not support this option. However I've never seen any
        // reports where this was causing issues. So for now, simply skip this
        // option, without returning that the filter was invalid.
      } else {
        throw new Error(`Unknown option '${option}' in filter '${text}'`);
      }
    }
  }

  // If no element types are mentioned, the default set is implied.
  // Otherwise, the element types are used, which can be ElementTypes.NONE
  if (allowedElementTypes === undefined) {
    result.allowedElementTypes = ElementTypes.DEFAULTTYPES;
  } else {
    result.allowedElementTypes = allowedElementTypes;
  }

  // We parse whitelist rules too, in which case we already know it's a
  // whitelist rule so can ignore the @@s.
  if (Filter.isWhitelistFilter(rule)) {
    rule = rule.substring(2);
  }

  // Convert regexy stuff.

  // First, check if the rule itself is in regex form.  If so, we're done.
  const matchcase = (result.options & FilterOptions.MATCHCASE) ? '' : 'i';
  if (/^\/.+\/$/.test(rule)) {
    result.rule = rule.substr(1, rule.length - 2); // remove slashes
    result.rule = new RegExp(result.rule, matchcase);
    return result;
  }

  const key = rule.match(/[\w&=]{5,}/);
  if (key) {
    result.key = new RegExp(key, matchcase);
  }

  // ***** -> *
  // replace, excessive wildcard sequences with a single one
  rule = rule.replace(/\*-\*-\*-\*-\*/g, '*');

  rule = rule.replace(/\*\*+/g, '*');

  // Some chars in regexes mean something special; escape it always.
  // Escaped characters are also faster.
  // - Do not escape a-z A-Z 0-9 and _ because they can't be escaped
  // - Do not escape | ^ and * because they are handled below.
  rule = rule.replace(/([^a-zA-Z0-9_|^*])/g, '\\$1');

  // ^ is a separator char in ABP
  rule = rule.replace(/\^/g, '[^\\-\\.\\%a-zA-Z0-9_]');

  // If a rule contains *, replace that by .*
  rule = rule.replace(/\*/g, '.*');

  // Starting with || means it should start at a domain or subdomain name, so
  // match ://<the rule> or ://some.domains.here.and.then.<the rule>
  rule = rule.replace(/^\|\|/, '^[^\\/]+\\:\\/\\/([^\\/]+\\.)?');

  // Starting with | means it should be at the beginning of the URL.
  rule = rule.replace(/^\|/, '^');

  // Rules ending in | means the URL should end there
  rule = rule.replace(/\|$/, '$');

  // Any other '|' within a string should really be a pipe.
  rule = rule.replace(/\|/g, '\\|');

  // If it starts or ends with *, strip that -- it's a no-op.
  rule = rule.replace(/^\.\*/, '');
  rule = rule.replace(/\.\*$/, '');

  result.rule = new RegExp(rule, matchcase);
  return result;
};

// Blocking and whitelist rules both become PatternFilters.
PatternFilter.prototype = {
  // Inherit from Filter.
  __proto__: Filter.prototype,

  // Returns true if an element of the given type loaded from the given URL
  // would be matched by this filter.
  //   url:string the url the element is loading.
  //   elementType:ElementTypes the type of DOM element.
  //   isThirdParty: true if the request for url was from a page of a
  //       different origin
  matches(url, elementType, isThirdParty) {
    if (!(elementType & this.allowedElementTypes)) {
      return false;
    }

    // If the resource is being loaded from the same origin as the document,
    // and your rule applies to third-party loads only, we don't care what
    // regex your rule contains, even if it's for someotherserver.com.
    if ((this.options & FilterOptions.THIRDPARTY) && !isThirdParty) {
      return false;
    }

    if ((this.options & FilterOptions.FIRSTPARTY) && isThirdParty) {
      return false;
    }

    if (this.key && !this.key.test(url)) {
      return false;
    }

    return this.rule.test(url);
  },

  /**
   * Rewrites an URL.
   * @param {string} url the URL to rewrite
   * @return {string} the rewritten URL, or the original in case of failure
   */
  rewriteUrl(url) {
    try {
      const rewrittenUrl = new URL(url.replace(this.rule, this.rewrite), url);
      if (rewrittenUrl.origin === new URL(url).origin) {
        return rewrittenUrl.href;
      }
    } catch (e) {
      // silently ignore any exceptions
    }

    return url;
  },
};

ContentFilter.prototype = {
  // Inherit from Filter.
  __proto__: Filter.prototype,

  /**
   * The body of the filter
   * @type {string}
   */
  body: null,
};

/**
 * Creates a content filter from a pre-parsed text representation
 *
 * @param {string} text
 * @param {string} [domains]
 *   domains part of the text representation
 * @param {string} [type]
 *   rule type: "$" for a snippet filter
 * @param {string} body
 *   body part of the text representation, a snippet script
 * @return {SnippetFilter}
 */
ContentFilter.fromText = function contentFilterFromText(text, domains, type, body) {
  let localDomains = domains;
  let localType = type;
  let localBody = body;
  if (!localDomains && !localType && !localBody && Filter.isContentFilter(text)) {
    const match = Filter.contentRegExp.exec(text);
    [, localDomains, localType, localBody] = match;
  }
  // We don't allow content filters which have any empty domains.
  if (localDomains && /(^|,)~?(,|$)/.test(localDomains)) {
    throw new Error(`Invalid Content Filter text: ${text} domain: ${localDomains} - no domain(s)`);
  }

  if (localType === '$') {
    return new SnippetFilter(text, localDomains, localBody);
  }
  throw new Error(`Invalid Content Filter text: ${text} type: ${localType}`);
};

SnippetFilter.prototype = {
  // Inherit from Filter.
  __proto__: Filter.prototype,

  /**
   * Script that should be executed
   * @type {string}
   */
  get script() {
    return this.body;
  },
};

// If !|excludeFilters|, returns filter.
// Otherwise, returns a new SnippetFilter that is the combination of
// |filter| and each selector exclusion filter in the given list.
SnippetFilter.merge = function snippetFilterMerge(filter, excludeFilters) {
  if (!excludeFilters) {
    return filter;
  }

  const domains = filter.domains.clone();
  let excludeFiltersLength = excludeFilters.length;
  while (excludeFiltersLength > 0) {
    excludeFiltersLength -= 1;
    domains.subtract(excludeFilters[excludeFiltersLength].domains);
  }

  return new SnippetFilter(filter.text, domains, filter.body);
};

// Attach object to window
Object.assign(window, {
  Filter,
  SelectorFilter,
  ElemHideEmulationFilter,
  PatternFilter,
  SnippetFilter,
});
